# -*- encoding: utf-8 -*-
# certara.py - Data handling routines developed by Optibrium for intregration with StarDrop software
# Use of this code is subject to the terms of the StarDrop License Agreement and use, copying or redistribution of this file for use by other persons or organisations is prohibited
# Copyright(C) Optibrium Ltd 2016
# stardrop-support@optibrium.com
#

import app
import string
import time
import sys

try:
	import json
except:
	import simplejson as json
	
import urllib2
import base64

import PythonQt

import pickle

from certaraconfig import *

try:
	import refreshscript

	def refreshD360(method,params):
		print 'refreshing D360'
		ds = loadData(int(method))
		return ds.pack()

	refreshscript.setRefreshFunction(refreshD360)
except:
	pass


user = ''
pwd = ''

def getCredentials():
	global user
	global pwd
	if user == '' and pwd == '':
		login = PythonQt.LogInDialog()
		if login.run():
			user = login.getUser()
			pwd = login.getPassword()
			
def handleConnectionError():
	global user
	global pwd
	user = ''
	pwd = ''
	raise Exception("Connection to D360 failed")
	
def getResultFromServer(url):
	global user
	global pwd
	getCredentials()
	print url
	base64string = base64.encodestring('%s:%s' % (user, pwd)).replace('\n', '')	
	query = urllib2.Request(url)
	query.add_header("Authorization", "Basic %s" % base64string) 

	try:
		resp = urllib2.urlopen(query).read()
	except:
		print "Connection error ", sys.exc_info()[0]
		handleConnectionError()
	return resp
	
def splitWithQuotes(input, delim):
	stringList = []
	next = ''
	inQuotes = False
	for inputChar in input:
		if inputChar == '"':
			if inQuotes:
				inQuotes = False
			else:
				inQuotes = True
			continue
		if inputChar == delim and not inQuotes:
			stringList.append(next)
			next = ''
		else:
			next = next + inputChar
	stringList.append(next)
	return stringList

	
def getReports(id,prefix,name):
	reports = []
	url = URL_BASE + "workspace/listItems/" + `id`
	details = json.loads(getResultFromServer(url))
		
	if details.has_key("childItems"):
		for ds in details["childItems"]:
			if ds["type"] == "RESULT_SET":
				qualifiedName = prefix + ds["name"]
				reports.append([qualifiedName,int(ds["id"])])
	if details.has_key("childFolders"):
		for folder in details["childFolders"]:
			newPrefix =  prefix  + folder["name"] + "/"
			newReports = getReports(folder["id"],newPrefix,name)
			reports.extend(newReports)
	return reports
	
def getWorkspaceList():
	print 'refreshing list'
	url = URL_BASE + "workspace/list"
	workspaces = json.loads(getResultFromServer(url))
	reports = {}
	for wsp in workspaces:
		id = wsp["id"]
		name = wsp["displayName"]
		reportsForWorkspace = getReports(id,"", name)
		if len(reportsForWorkspace) > 0:
			reports[name] = reportsForWorkspace
	return reports
	
def refreshWorkspaceList():
	reports = getWorkspaceList()
	try:
		output = open(WORKSPACE_CACHE,"wb")
		pickle.dump(reports,output)
		output.close()
	except:
		print "Failed to write file"
	
		
def getAvailableReports():
	try:
		input = open(WORKSPACE_CACHE,"rb")
		reports = pickle.load(input)
		input.close()
		return reports
	except:
		return getWorkspaceList()
	
def checkType(values,name):
	#check for molecules/structures
	rv = app.cont
	if name.upper().find("STRUCT") != -1 or name.upper().find("SMILES") != -1:
		rv = app.mol
		return rv
	#check for numerical values - check first TYPE_CHECK_DEPTH, and look for at least TYPE_CHECK_FRACTION to be numbers
	count = 0.0
	numbers = 0.0
	index = 0
	while count < TYPE_CHECK_DEPTH:
		if index < len(values):
			#check for a null - count as valid number
			if values[index] is None or values[index] == '':
				pass
			else:
				count += 1.0
				try:
					v = float(values[index])
					numbers += 1.0
				except:
					if values[index].lower() == 'invalid':#count this entry as representing a number column with invalid entries
						numbers += 1.0
					elif len(values[index]) > 0 and values[index][0] in ('<','>','='):#count a EQ, GT or LT operator as being a number
						numbers += 1.0
					pass
			index += 1
		else:
			break
	if count > 0:
		if numbers/count >= TYPE_CHECK_FRACTION:
			rv = app.cont
		else:
			rv = app.string
	return rv
	
def extractResult(result):
	#if the data is empty or invalid at this point an exception will be thrown thrown and caught in loadData
	qual = str(0)
	value = 0.0
	main = result.rstrip(string.letters)
	if main[0] == '<':
		qual = '\x01'
		main = main.lstrip('<')
		main = main.lstrip()
	elif main[0] == '>':
		qual = '\x02'
		main = main.lstrip('>')
		main = main.lstrip()
	elif main[0] == '=':
		main = main.lstrip('=')
		main = main.lstrip()
	value = float(main)
	return qual, value
	
def extractValueFromList(results):
	count = 0
	total = 0.0
	for r in results:
		try:
			value = float(r.lstrip().rstrip())
			total += value
			count += 1
		except:
			pass
	if count > 0:
		return app.Dataset.cont(total/count,0.0)
	else:
		return app.Dataset.createInvalidEntry(app.cont)
	
def createContEntry(entry):
	value = 0.0
	error = 0.0
	qualifier = str(0)
	try:
		value = float(entry)
	except:
		results = entry.split(',')
		if len(results) > 1:
			#print "Entry contains a list"
			return extractValueFromList(results)
		qualifier, value = extractResult(results[0].lstrip().rstrip())
		pass
	return (12+2*8+1, (True, [], []), (qualifier, value,error))
	


def loadData(report):
	ds = app.Dataset()	
	url = URL_BASE + "dataset/" + `report` + "/csv"

	entries = getResultFromServer(url).split("\n")
	headers = splitWithQuotes(entries[0].rstrip(),',')
	if len(headers) > 0:
		table = []
		for h in headers:
			table.append([])


		#now iterate over the rows and build an array of the entries
		for row in entries[1:]:
			if len(row) == 0:
				continue
			data = splitWithQuotes(row.rstrip(),',')
			if len(data) > len(headers):
				print row
				raise Exception('Incomplete data')
			elif len(data) < len(headers):
				while len(data) < len(headers):
					data.append('')
			#add data for 'normal' columns
			for c in xrange(len(headers)):
				table[c].append(data[c].rstrip('"').lstrip('"'))
		#now build the columns from the data in the table
		ds.cols = []
		for i,header in enumerate(headers):
			col = []
			if len(table[i]) > 0:
				type = checkType(table[i],header)
				details = ''
				try:
					if type == app.cont:
						details = ((0.0,0.0),(0,0))
					ds.columns.append(app.Dataset.meta(header.rstrip('"').lstrip('"').decode('latin-1'),type,details))
					for entry in table[i]:
						if entry == None:
							entry = ''
						if type == app.mol:
							col.append(app.Dataset.molecule(entry))
						elif type == app.string:
							col.append(app.Dataset.string(entry))
						else:
							try:
								col.append(createContEntry(entry))
							except Exception,e:
								col.append(app.Dataset.createInvalidEntry(app.cont))
								if str(e).find("Entry contains a list") != -1:
									raise
				except Exception,e:
					#add as a string column instead
					ds.columns.pop()
					details = ''
					col = []
					ds.columns.append(app.Dataset.meta(header.rstrip('"').lstrip('"').decode('latin-1'),app.string,details))
					for entry in table[i]:
						if entry == None:
							entry = ''
						col.append(app.Dataset.string(entry))
						
						
			else:
				ds.columns.append(app.Dataset.meta(header,app.string,''))
			ds.cols.append(col)
	else:
		raise(Exception('File contains no data'))
	return ds
